//: $CC -DSP_LIB -DLS_LIB -c $file $OPTIONS
#ifndef SP_LDLT_DIAGONALC
#define SP_LDLT_DIAGONALC
#include "sp_names.h"
#include SP_LDLTDIAGONAL_H

sp_LDLT_diagonal::sp_LDLT_diagonal(char *nam):eps(ls_EPS)
{ strcpy(name=ONAME,nam);
  A_name=A.name; F_name=F.name; ind_name=ind.name;
  strcpy(F.name,"LDLT_diagonal-Factorization");
  strcpy(ind.name,"Diagonal-Indices"); rc=-1;
}

sp_LDLT_diagonal::sp_LDLT_diagonal(sp_LDLT_diagonal &B)
{ eps=B.eps; strcpy(name,B.name); *(B.name)=0;
  A_name=A.name; F_name=F.name; ind_name=ind.name; rc=B.rc;
}

sp_LDLT_diagonal& sp_LDLT_diagonal::swap(sp_LDLT_diagonal &B)
{ char na[ls_len]; strcpy(na,name); strcpy(name,B.name); strcpy(B.name,na);
  eps=B.eps; A.swap(B.A); F.swap(B.F); ind.swap(B.ind);
  int r=rc; rc=B.rc, B.rc=r;
  return *this;
}

const sp_LDLT_diagonal& sp_LDLT_diagonal::operator=(const sp_LDLT_diagonal &B)
{ eps=B.eps; A=B.A; F=B.F; ind=B.ind; rc=B.rc; return *this;}

sp_LDLT_diagonal::sp_LDLT_diagonal(sp_sMatrix &B,ls_REAL alpha,char *nam):\
eps(ls_EPS)
{ static char *q1="sp_LDLT::sp_LDLT_diagonal";
  A.swap(B); strcpy(name=ONAME,nam);
  A_name=A.name; F_name=F.name; ind_name=ind.name;
  ls_UINT i, j, k, l, n=A.dimension();
  ls_array1<ls_UINT> kind(n);
  ind.swap(kind); strcpy(ind.name,"Diagonal-Indices");
  F=A; strcpy(F.name,"LDLT_diagonal-Factorization");
  ls_UINT *in=ind.asArray();
  double s, t;
  list2<ls_REAL> *aa, *ap, *ak, *aki, *akj, *c,
        *anf=F.asList(), *a=anf->r_next;
  for (i=0; i<n; in[i]=i, i++);
  if(alpha>0) for(k=0; k < n; s=F.get(k,k), F.put(s+alpha,k,k), k++);
  for (i=0; i<n; i++)
  { for (j=i+1, k=i, l=(a+i)->i+(a+i)->j; j<n; j++)
    if((a+j)->i+(a+j)->j<l) l=(a+j)->i+(a+j)->j, k=j;
    if(i<k) F.swap_row(i,k), l=in[i], in[i]=in[k], in[k]=l;
  }
  for (k=0; k < n-1; k++)
  { ak=(a+k)->r_next;
    if ((ak->j!=k)||(fabs(t=ak->x)<eps))
    { rc=k+1; F.remove(); ind.remove(); return;}
    t=1/t; aki=ak;
    while ((aki=aki->r_next)!=anf)
    { s=aki->x*t, akj=aki, i=aki->j;
      while (akj->j<n)
      { j=akj->j; ap=a+i;
        while ((aa=ap->r_next)->j < j) ap=aa;
        if (aa->j==j) aa->x=ls_REAL(aa->x-s*akj->x);
        else
        { ap->r_next=c=F.get2_el(); c->r_next=aa,
          c->x=ls_REAL(-s*akj->x), c->j=j, c->i=i;
          ap=a+j; while((aa=ap->c_next)->i<i) ap=aa;
          c->c_next=aa, ap->c_next=c,
          (a+j)->j+=1, (a+i)->i+=1;
        }
        akj=akj->r_next;
      }
      aki->x=ls_REAL(s);
    }
  }
  rc=0;
}

const sp_LDLT_diagonal& sp_LDLT_diagonal::\
solve(ls_Vector &x, const ls_Vector &b) const
{ static char *q1="sp_LDLT_diagonal::solve";
  ls_UINT n=F.dimension();
  if(!n) throw ls_error(name,q1,no_factorization_given_);
  list2<ls_REAL> *aa, *anf=F.asList(), *a=anf->r_next;
  ls_UINT k, *in=ind.asArray();
  if(b.dimension()!=n) throw ls_error(b.name,q1,size_error_);
  ls_REAL *xx=x.asArray(), *bb=b.asArray();
  double s; x=b;
  for(k=0; k<n; xx[k]=bb[in[k]], k++);
  for(k=0; k<n-1; k++)
  { aa=(a+k)->r_next, s=xx[k];
    while((aa=aa->r_next)!=anf) xx[aa->j]=ls_REAL(xx[aa->j]- s*aa->x);
  }
  for (k=0; k<n; xx[k]/=((a+k)->r_next)->x, k++);
  while(--n)
  { for (k=0, s=xx[n]; k<n; k++)
    { aa=a+k; while((aa=aa->r_next)->j<n);
      if(aa->j==n) xx[aa->i]=ls_REAL(xx[aa->i]-s*aa->x);
    }
  }
  return *this;
}

ls_UINT sp_LDLT_diagonal::post_iteration(ls_Vector &x, ls_Vector &b) const
{ static char *q1="sp_LDLT::post_iteration";
  list2<ls_REAL> *anf=F.asList(), *a=anf->r_next;
  if(sizeof(ls_REAL)<8) throw ls_error(name,q1,set_ls_REAL_to_double_);
  ls_UINT rep=10, l=0, i, n=anf->i;
  if(!n) throw ls_error("",q1,no_factorization_given_);
  ls_Vector v(n), h(n);
  ls_REAL *vv=v.asArray(), *hh=h.asArray(), *xx=x.asArray();
  for(i=0; i<n; hh[i]=xx[i]-(vv[i]=float(xx[i])), i++);
  ls_Vector r, z(n); r=(b-A*v)-A*h;
  while((l++<rep)&&!(r*r<eps))
  { solve(z,r); x+=z;
    for(i=0; i<n; hh[i]=xx[i]-(vv[i]=float(xx[i])), i++);
    r=(b-A*v)-A*h;
  }
  b=r; return l-1;
}

ls_Vector sp_LDLT_diagonal::residuum(const ls_Vector &x, const ls_Vector &b) const
{  ls_Vector r(0,"Residuum"); r=b-A*x; return r;}

const sp_LDLT_diagonal& sp_LDLT_diagonal::write_row(ostream &f) const
{ f<<name<<endl<<"{"<<endl;
  A.write_row(f), F.write_row(f); ind.write(f); f<<"}"<<endl;
  return *this;
}

sp_LDLT_diagonal& sp_LDLT_diagonal::read_row(istream &f)
{ static char *q1="sp_LDLT_diagonal::read_row";
  char w[ls_len];
  if(*name)while(f>>w&&strcmp(w,name));
  else f>>name, *(A.name)=0, *(F.name)=0, *(ind.name)=0;
  if(!f.good()) throw ls_error(name, q1, object_not_found_);
  f>>w;
  if(strcmp(w, "{")) throw ls_error(name,q1,object_contents_error_);
  A.read_row(f), F.read_row(f), ind.read(f);
  if(!(f>>w)||(*w!='}')) throw ls_error(name,q1,object_end_error_);
  rc=0; return *this;
}

const sp_LDLT_diagonal& sp_LDLT_diagonal::write_column(ostream &f) const
{ f<<name<<endl<<"{"<<endl;
  A.write_column(f), F.write_column(f), ind.write(f); f<<"}"<<endl;
  return *this;
}

sp_LDLT_diagonal& sp_LDLT_diagonal::read_column(istream &f)
{ static char *q1="sp_LDLT_diagonal::read_column";
  char w[ls_len];
  if(*name)while(f>>w&&strcmp(w,name));
  else f>>name, *(A.name)=0, *(F.name)=0, *(ind.name)=0;
  if(!f.good()) throw ls_error(name, q1, object_not_found_);
  f>>w;
  if(strcmp(w, "{")) throw ls_error(name,q1,object_contents_error_);
  A.read_column(f), F.read_column(f), ind.read(f);
  if(!(f>>w)||(*w!='}'))
    throw ls_error(name, q1, object_end_error_);
  rc=0; return *this;
}

const sp_LDLT_diagonal& sp_LDLT_diagonal::operator>>(char *nam) const
{ if(!*nam)
  { cout.setf(ios::fixed, ios::floatfield);
    cout.precision(int(sizeof(ls_REAL)*1.5));
    write_row(cout);
  }
  else
  { ofstream f(nam);
    f.setf(ios::scientific, ios::floatfield);
    f.precision(int(sizeof(ls_REAL)*1.5));
    write_row(f);
  }
  return *this;
}

sp_LDLT_diagonal& sp_LDLT_diagonal::operator<<(char *nam)
{ if(!*nam) read_row(cin); else{ ifstream f(nam); read_row(f);}
  return *this;
}
#endif
