//: $CC -DSP_LIB -DLS_LIB -c $file $OPTIONS
#ifndef SP_LU_DIAGONALC
#define SP_LU_DIAGONALC
#include "sp_names.h"
#include SP_LUDIAGONAL_H

sp_LU_diagonal::sp_LU_diagonal(char *nam):eps(ls_EPS)
{ strcpy(name=ONAME,nam); A_name=A.name; F_name=F.name; ind_name=ind.name;
  strcpy(F.name,"LU_diagonal-Factorization");
  strcpy(ind.name,"Diagonal-Indices"); rc=-1;
}

sp_LU_diagonal::sp_LU_diagonal(sp_LU_diagonal &B)
{ eps=B.eps; strcpy(name,B.name); *(B.name)=0; rc=B.rc;
  A_name=A.name; F_name=F.name; ind_name=ind.name;
}

sp_LU_diagonal& sp_LU_diagonal::swap(sp_LU_diagonal &B)
{ char na[ls_len]; strcpy(na,name); strcpy(name,B.name); strcpy(B.name,na);
  eps=B.eps; A.swap(B.A); F.swap(B.F); ind.swap(B.ind);
  int r=rc; rc=B.rc, B.rc=r;
  return *this;
}

const sp_LU_diagonal& sp_LU_diagonal::operator=(const sp_LU_diagonal &B)
{ eps=B.eps; A=B.A; F=B.F; ind=B.ind; rc=B.rc;return *this;}

sp_LU_diagonal::sp_LU_diagonal(sp_Matrix &B, ls_REAL alpha, char *nam):eps(ls_EPS)
{ static char *q1="sp_LU::sp_LU_diagonal";
  A.swap(B); strcpy(name=ONAME,nam);
  A_name=A.name; F_name=F.name; ind_name=ind.name;
  ls_UINT i, j, k, is, l, n=A.number_of_rows();
  if(n!=A.number_of_columns())
    throw ls_error(B.name,q1,object_not_quadratic_);
  F=A; strcpy(F.name,"LDLT_diagonal-Factorization");
  ls_array1<ls_UINT> kind(n);
  ind.swap(kind); strcpy(ind.name,"Diagonal-Indices");
  list2<ls_REAL> *aa, *ap, *ad, *ai, *aip, *aj, *ajp, *aij, *c,
         *anf=F.asList(), *a=anf->r_next;
  double s, ss;
  for (i=0; i<n; ind[i]=i, i++);
  if(alpha>0) for(k=0; k < n; s=F.get(k,k), F.put(s+alpha,k,k), k++);
  for (i=0; i<n; i++)
  { ad=a+i; while((ad=ad->r_next)->j<i);
    if(ad->j==i) s=fabs(ad->x), is=i; else s=0, is=0;
    for (k=i+1; k<n; k++)
    { ai=a+k; while((ai=ai->r_next)->j<k);
      if((ai->j==k)&&(ss=fabs(ai->x))>s) s=ss, is=k;
    }
    if(s<eps){ rc=i+1; F.remove(); ind.remove(); return;}
    if(i<is)
    { F.swap_row(i, is), F.swap_column(i, is),
      l=ind[i], ind[i]=ind[is], ind[is]=l;
      ad=a+i; while((ad=ad->r_next)->j<i);
    }
    s=1/ad->x; aj=a+i; while((aj=aj->c_next)->i<i);
    if(ad->r_next==anf)
    { while((aj=aj->c_next)!=anf) aj->x=ls_REAL(aj->x*s); continue;}
    ajp=aj;
    while((aj=ajp->c_next)!=anf)
    { if(fabs(ss=aj->x*s)<eps)
      { ajp->c_next=aj->c_next,
        ap=a+aj->i, ap->i-= 1, (a+aj->j)->j-= 1;
        while((aa=ap->r_next)!=aj) ap=aa;
        ap->r_next=aa->r_next; F.ret2_el(aj);
        continue;
      }
      aj->x=ls_REAL(ss), ajp=aj, aij=ad, k=aj->i, aip=a+k;
      while((aij=aij->r_next)!=anf)
      { j=aij->j;
        while((ai=aip->r_next)->j<j) aip=ai;
        if(ai->j==j) { ai->x=ls_REAL(ai->x-ss*aij->x); continue; }
        ajp=a+j, aj->j+=1, (a+k)->i+=1;
        while((aj=ajp->c_next)->i<k) ajp=aj;
        c=F.get2_el(); c->x=ls_REAL(-ss*aij->x), c->r_next=ai,
        c->c_next=aj, aip->r_next=ajp->c_next=c;
      }
    }
  }
  rc=0;
}

const sp_LU_diagonal& sp_LU_diagonal::\
solve(ls_Vector &x, const ls_Vector &b) const
{ static char *q1="sp_LU_diagonal::solve";
  ls_UINT n=F.number_of_rows();
  if(!n) throw ls_error(name,q1,no_factorization_given_);
  list2<ls_REAL> *aa, *anf=F.asList(), *a=anf->r_next;
  ls_UINT i, k, *in=ind.asArray();
  if(b.dimension()!=n) throw ls_error(b.name,q1,size_error_);
  double s, ss, sp, sm;
  ls_REAL *xx=x.asArray(), *bb=b.asArray();
  for (i=0; i<n; xx[in[i]]=ls_REAL(bb[in[i]]-(sp+sm)), i++)
    for (sp=sm=0, aa=a+i; (k=(aa=aa->r_next)->j)<i;)
      if((s=xx[in[k]]*aa->x)<0) sm+=s; else sp+=s;
  for (i=n; i--; xx[in[i]]=ls_REAL((xx[in[i]]-(sp+sm))/ss))
  { aa=a+i; while((aa=aa->r_next)->j<i);
    for (sm=sp=0, ss=aa->x; (k=(aa=aa->r_next)->j) < n;)
      if((s=xx[in[k]]*aa->x)<0) sm+=s; else sp+=s;
  }
  return *this;
}

ls_UINT sp_LU_diagonal::post_iteration(ls_Vector &x, ls_Vector &b) const
{ static char *q1="sp_LU_diagonal::post_iteration";
  if(sizeof(ls_REAL)<8) throw ls_error(name,q1,set_ls_REAL_to_double_);
  list2<ls_REAL> *anf=F.asList();
  ls_UINT rep=10, l=0, i, n=anf->i;
  if(!n) throw ls_error(name,q1,no_factorization_given_);
  ls_Vector v(n), h(n);
  ls_REAL *vv=v.asArray(), *hh=h.asArray(), *xx=x.asArray();
  for(i=0; i<n; hh[i]=xx[i]-(vv[i]=float(xx[i])), i++);
  ls_Vector r, z(n); r=(b-A*v)-A*h;
  while((l++<rep)&&!(r*r<eps))
  { solve(z,r); x+=z;
    for(i=0; i<n; hh[i]=xx[i]-(vv[i]=float(xx[i])), i++);
    r=(b-A*v)-A*h;
  }
  b=r; return l-1;
}

ls_Vector sp_LU_diagonal::residuum(const ls_Vector &x, const ls_Vector &b) const
{  ls_Vector r(0,"Residuum"); r=b-A*x; return r;}

const sp_LU_diagonal& sp_LU_diagonal::write_row(ostream &f) const
{ f<<name<<endl<<"{"<<endl;
  A.write_row(f), F.write_row(f), ind.write(f); f<<"}"<<endl;
  return *this;
}

const sp_LU_diagonal& sp_LU_diagonal::write_column(ostream &f) const
{ f<<name<<endl<<"{"<<endl;
  A.write_column(f), F.write_column(f), ind.write(f); f<<"}"<<endl;
  return *this;
}

sp_LU_diagonal& sp_LU_diagonal::read_row(istream &f)
{ static char *q1="sp_LU_diagonal::read_row";
  char w[ls_len];
  if(*name)while(f>>w&&strcmp(w,name));
  else f>>name, *(A.name)=0, *(F.name)=0, *(ind.name)=0;
  if(!f.good()) throw ls_error(name, q1, object_not_found_);
  f>>w;
  if(strcmp(w, "{")) throw ls_error(name,q1,object_contents_error_);
  A.read_row(f), F.read_row(f), ind.read(f);
  if(!(f>>w)||(*w!='}')) throw ls_error(name, q1, object_end_error_);
  rc=0; return *this;
}

sp_LU_diagonal& sp_LU_diagonal::read_column(istream &f)
{ static char *q1="sp_LU_diagonal::read_diagonal";
  char w[ls_len];
  if(*name)while(f>>w&&strcmp(w,name));
  else f>>name, *(A.name)=0, *(F.name)=0, *(ind.name)=0;
  if(!f.good()) throw ls_error(name, q1, object_not_found_);
  f>>w;
  if(strcmp(w, "{")) throw ls_error(name,q1,object_contents_error_);
  A.read_column(f), F.read_column(f), ind.read(f);
  if(!(f>>w)||(*w!='}')) throw ls_error(name, q1, object_end_error_);
  rc=0; return *this;
}

const sp_LU_diagonal& sp_LU_diagonal::operator>>(char *nam) const
{ if(!*nam)
  { cout.setf(ios::fixed, ios::floatfield);
    cout.precision(int(sizeof(ls_REAL)*1.5));
    write_row(cout);
  }
  else
  { ofstream f(nam);
    f.setf(ios::scientific, ios::floatfield);
    f.precision(int(sizeof(ls_REAL)*1.5));
    write_row(f);
  }
  return *this;
}

sp_LU_diagonal& sp_LU_diagonal::operator<<(char *nam)
{ if(!*nam) read_row(cin); else{ ifstream f(nam); read_row(f);}
  return *this;
}
#endif
