//: $CC -c $file $OPTIONS
#ifndef SP_MATRIXC
#define SP_MATRIXC
#include "sp_names.h"
#include SP_MATRIX_H

sp_Matrix::sp_Matrix(ls_UINT m, ls_UINT n, char *nam):
sp_list2<ls_REAL>(m, n,nam){}

sp_Matrix::sp_Matrix(ls_REAL *B, ls_UINT m, ls_UINT n, char *nam):
sp_list2<ls_REAL>(B,m,n,nam){}

sp_Matrix::sp_Matrix(sp_Matrix &B):sp_list2<ls_REAL>(B){}

const sp_Matrix& sp_Matrix::operator=(const sp_Matrix &B)
{ sp_list2<ls_REAL>::operator=(B); return *this;}

sp_Matrix sp_Matrix::operator+(const sp_Matrix &B)const
{ static char *q1="sp_Matrix::operator+";
  ls_UINT m=B.number_of_rows(), n=B.number_of_columns();
  if((m!=a->i)||(n!=a->j)) throw ls_error(name,q1,size_error_);
  sp_Matrix C(m,n);
  list2<ls_REAL> *b=B.asList(), *aa=a->r_next, *bb=b->r_next, *an, *bn;
  ls_UINT i;
  for(i=0; i<m; i++, aa++, bb++)
  { an=aa->r_next; bn=bb->r_next;
    while(an!=a)
      if(an->j<bn->j) C.put(an->x,i,an->j), an=an->r_next;
      else
        if(an->j>bn->j) C.put(bn->x,i,bn->j), bn=bn->r_next;
        else C.put(an->x+bn->x,i,an->j), an=an->r_next, bn=bn->r_next;
    while(bn!=b) C.put(bn->x,i,bn->j), bn=bn->r_next;
  }
  return C;
}

sp_Matrix sp_Matrix::operator-(const sp_Matrix &B)const
{ static char *q1="sp_Matrix::operator-";
  ls_UINT m=B.number_of_rows(), n=B.number_of_columns();
  if((m!=a->i)||(n!=a->j)) throw ls_error(name,q1,size_error_);
  sp_Matrix C(m,n);
  list2<ls_REAL> *b=B.asList(), *aa=a->r_next, *bb=b->r_next, *an, *bn;
  ls_UINT i;
  for(i=0; i<m; i++, aa++, bb++)
  { an=aa->r_next; bn=bb->r_next;
    while(an!=a)
      if(an->j<bn->j) C.put(an->x,i,an->j), an=an->r_next;
      else
        if(an->j>bn->j) C.put(bn->x,i,bn->j), bn=bn->r_next;
        else C.put(an->x-bn->x,i,an->j), an=an->r_next, bn=bn->r_next;
    while(bn!=b) C.put(bn->x,i,bn->j), bn=bn->r_next;
  }
  return C;
}

sp_Matrix& sp_Matrix::operator+=(const sp_Matrix &B)
{ static char *q1="sp_Matrix::operator+=";
  ls_UINT m=B.number_of_rows(), n=B.number_of_columns();
  if((m!=a->i)||(n!=a->j)) throw ls_error(name,q1,size_error_);
  sp_Matrix C(m,n);
  list2<ls_REAL> *b=B.asList(), *aa=a->r_next, *bb=b->r_next, *an, *bn;
  ls_UINT i;
  for(i=0; i<m; i++, aa++, bb++)
  { an=aa->r_next; bn=bb->r_next;
    while(an!=a)
      if(an->j<bn->j) C.put(an->x,i,an->j), an=an->r_next;
      else
        if(an->j>bn->j) C.put(bn->x,i,bn->j), bn=bn->r_next;
        else C.put(an->x+bn->x,i,an->j), an=an->r_next, bn=bn->r_next;
    while(bn!=b) C.put(bn->x,i,bn->j), bn=bn->r_next;
  }
  swap(C); C.remove();
  return *this;
}

sp_Matrix& sp_Matrix::operator-=(const sp_Matrix &B)
{ static char *q1="sp_Matrix::operator-=";
  ls_UINT m=B.number_of_rows(), n=B.number_of_columns();
  if((m!=a->i)||(n!=a->j)) throw ls_error(name,q1,size_error_);
  sp_Matrix C(m,n);
  list2<ls_REAL> *b=B.asList(), *aa=a->r_next, *bb=b->r_next, *an, *bn;
  ls_UINT i;
  for(i=0; i<m; i++, aa++, bb++)
  { an=aa->r_next; bn=bb->r_next;
    while(an!=a)
      if(an->j<bn->j) C.put(an->x,i,an->j), an=an->r_next;
      else
        if(an->j>bn->j) C.put(bn->x,i,bn->j), bn=bn->r_next;
        else C.put(an->x-bn->x,i,an->j), an=an->r_next, bn=bn->r_next;
    while(bn!=b) C.put(bn->x,i,bn->j), bn=bn->r_next;
  }
  swap(C); C.remove();
  return *this;
}

sp_Matrix sp_Matrix::operator*(const sp_Matrix &B)const
{ static char *q1="sp_Matrix::operator*";
  ls_UINT m=B.number_of_rows(), n=B.number_of_columns();
  if(m!=a->j) throw ls_error(name,q1,size_error_);
  sp_Matrix C(a->i,n); ls_UINT k;
  list2<ls_REAL> *c=C.asList(), *c_anf=c->c_next,
                 *c_m=c_anf+a->i, *c_n=c_anf+n, *c_i, *c_ij, *c_j, *c_p,
                 *a_anf=a->r_next, *a_m=a_anf+a->i, *a_l=a_anf+a->j,
                 *a_i, *a_ik,
                 *b=B.asList(), *b_anf=b->c_next,
                 *b_l=b_anf+m, *b_n=b_anf+n, *b_j, *b_kj;
  double sp, sm, s;
  for(a_i=a_anf, c_i=c_anf; c_i<c_m; c_ij->r_next=c, c_i++, a_i++)
    for(c_ij=c_i, b_j=b_anf; b_j<b_n; b_j++)
    { sp=sm=0; a_ik=a_i; b_kj=b_j->c_next;
      while((a_ik=a_ik->r_next)!=a)
      { k=a_ik->j;
        while(b_kj->i<k) b_kj=b_kj->c_next;
        if(b_kj->i==k)
        { if((s=b_kj->x*a_ik->x)<0) sm+=s; else sp+=s;
          b_kj=b_kj->c_next; continue;
        }
      }
      if(!(s=sp+sm)) continue;
      c_ij=c_ij->r_next=C.get2_el(); c_ij->i=(ls_UINT)(a_i-a_anf),
      c_ij->j=(ls_UINT)(b_j-b_anf), c_ij->x=ls_REAL(s), c_i->i+=1;
      c_p=c_anf+c_ij->j; c_p->j+=1;
      while((c_j=c_p->c_next)!=c) c_p=c_j;
      c_p->c_next=c_ij, c_ij->c_next=c;
    }
  return C;
}


ls_Vector sp_Matrix::operator* (const ls_Vector &v)const
{ static char *q1="sp_Matrix::operator*";
  ls_UINT n=a->j, m=a->i;
  if(v.dimension()!=n) throw ls_error(v.name,q1, size_error_);
  ls_Vector u(m);
  ls_REAL *uu=u.asArray(), *ue=uu+m, *vv=v.asArray();
  double s, sm, sp;
  list2<ls_REAL> *anf=a->r_next, *aa;
  while(uu<ue)
  { sp=sm=0, aa=anf;
    while((aa=aa->r_next)!=a)
      if((s=aa->x*vv[aa->j])<0) sm+=s; else sp+=s;
    *uu=ls_REAL(sp+sm); uu++, anf++;
  }
  return u;
}

sp_Vector sp_Matrix::operator* (const sp_Vector &v)const
{ static char *q1="sp_Matrix::*";
  ls_UINT n=a->j, m=a->i;
  if(v.dimension()!=n) throw ls_error(v.name,q1, size_error_);
  ls_UINT i=0, j;
  double s, sp, sm;
  sp_Vector u(m);
  list1<ls_REAL> *va=v.asList(), *vv, *vp, *ua=u.asList(), *uu=ua;
  list2<ls_REAL> *aa=a->r_next, *ae=aa+m, *c;
  while(aa<ae)
  { sp=sm=0, vp=va, c=aa;
    while((c=c->r_next)!=a)
    { j=c->j; while((vv=vp->next)->i<j) vp=vv;
      if(vv->i==j)
      { if((s=c->x*vv->x)<0) sm+=s; else sp+=s; vp=vv;}
      if(vv==va) break;
    }
    if(s=sp+sm)
    { uu=uu->next=u.get1_el(); uu->i=i, uu->x=ls_REAL(s), uu->next=ua;}
    aa++, i++;
  }
  return u;
}

ls_UINT sp_Matrix::solve(ls_Vector &x, ls_Vector &b) const
{ ls_UINT n=x.dimension();
  ls_Vector r, p; r=b-(*this)*x; p=r*(*this);
  ls_REAL s, rr=r*r, rrn, pp=p*p, eps=x.eps;
  ls_UINT it=0, rep=5*n;
  while (!((rr+pp)<eps)&&(it++<rep))
  { s=rr/pp, x+=s*p, r-=s*((*this)*p), rrn=r*r;
    p=r*(*this)+(rrn/rr)*p, pp=p*p, rr=rrn;
  }
  b=r; return it;
}

sp_Vector operator* (const sp_Vector &v,const sp_Matrix &B)
{ static char *q1="operator*";
  ls_UINT i=0, j, n=B.number_of_columns(), m=B.number_of_rows();
  if(v.dimension()!=m) throw ls_error(v.name,q1, size_error_);
  double s, sp, sm;
  sp_Vector u(n);
  list1<ls_REAL> *va=v.asList(), *vv, *vp, *ua=u.asList(), *uu=ua;
  list2<ls_REAL> *anf=B.asList(), *aa=anf->r_next, *ae=aa+n, *c;
  while(aa<ae)
  { sp=sm=0, vp=va, c=aa;
    while((c=c->c_next)!=anf)
    { j=c->j; while((vv=vp->next)->i<j) vp=vv;
      if(vv->i==j)
      { if((s=c->x*vv->x)<0) sm+=s; else sp+=s; vp=vv;}
      if(vv==va) break;
    }
    if(s=sp+sm)
    { uu=uu->next=u.get1_el(); uu->i=i, uu->x=ls_REAL(s), uu->next=ua;}
    aa++, i++;
  }
  return u;
}

ls_Vector operator* (const ls_Vector &v, const sp_Matrix &B)
{ static char *q1="operator*";
  ls_UINT n=B.number_of_columns(), m=B.number_of_rows();
  if(v.dimension()!=m) throw ls_error(v.name,q1, size_error_);
  ls_Vector u(n);
  ls_REAL *uu=u.asArray(), *ue=uu+n, *vv=v.asArray();
  double s, sp, sm;
  list2<ls_REAL> *anf=B.asList(), *aa, *ac=anf->c_next;
  while(uu<ue)
  { sp=sm=0, aa=ac;
    while((aa=aa->c_next)!=anf)
      if((s=aa->x*vv[aa->i])<0) sm+=s; else sp+=s;
    *uu=ls_REAL(sp+sm); uu++, ac++;
  }
  return u;
}
#endif
