//: $CC  -c $file $OPTIONS
#ifndef SP_sCG_PRECONDC
#define SP_sCG_PRECONDC
#include "sp_names.h"
#include SP_SCGPRECOND_H

sp_sCG_precond::sp_sCG_precond(char *nam):eps(ls_EPS)
{ strcpy(name=ONAME,nam); A_name=A.name; F_name=F.name;
  strcpy(F.name,"precond-Matrix"); rc=-1;
}

sp_sCG_precond::sp_sCG_precond(sp_sCG_precond &B)
{ eps=B.eps; strcpy(name,B.name); *(B.name)=0;
  A_name=A.name; F_name=F.name; rc=B.rc;
}

sp_sCG_precond& sp_sCG_precond::\
swap(sp_sCG_precond &B)
{ char na[ls_len]; strcpy(na,name); strcpy(name,B.name); strcpy(B.name,na);
  eps=B.eps; A.swap(B.A); F.swap(B.F);
  int r=rc; rc=B.rc, B.rc=r;
  return *this;
}

const sp_sCG_precond& sp_sCG_precond::\
operator=(const sp_sCG_precond &B)
{ eps=B.eps; A=B.A; F=B.F; rc=B.rc; return *this;}

sp_sCG_precond::sp_sCG_precond(sp_sMatrix &B,ls_REAL alpha,char *nam):
eps(ls_EPS)
{ static char *q1="sp_sCG_precond::sp_sCG_precond";
  A.swap(B); strcpy(name=ONAME,nam); A_name=A.name; F_name=F.name;
  ls_UINT i, j, k, n=A.dimension();
  double s, t; F=A; strcpy(F.name,"precond-Matrix");
  list2<ls_REAL> *aa, *ak, *aki, *akj, *anf=F.asList(), *a=anf->r_next;
  if(alpha>0) for(k=0; k < n; s=F.get(k,k), F.put(s+alpha,k,k), k++);
  for (k=0; k<n-1; k++)
  { ak=(a+k)->r_next;
    if ((ak->j!=k)||(fabs(t=ak->x) < eps)){ rc=k+1; F.remove(); return;}
    t=1/t; aki=ak;
    while ((aki=aki->r_next)!=anf)
    { s=aki->x*t, akj=aki, i=aki->j;
      while (akj->j<n)
      { j=akj->j; aa=a+i;
        while ((aa=aa->r_next)->j < j);
        if (aa->j==j) aa->x-=s*akj->x;
        akj=akj->r_next;
      }
      aki->x=s;
    }
  }
  rc=0;
}

ls_UINT sp_sCG_precond::solve(ls_Vector &x, ls_Vector &b) const
{ static char *q1="sp_sCG_precond::solve";
  list2<ls_REAL> *anf=F.asList(), *aa, *a=anf->r_next;
  ls_UINT l, k, n=a->j, rep=5*n, it=0;
  if(!n) throw ls_error(name,q1,no_factorization_given_);
  ls_Vector y(n), r, p(n); r=b-A*x; p=r;
  ls_REAL *pp=p.asArray(), *yy=y.asArray();
  double s;
// Loesen auf p
  for (k=0; k<n-1; k++)
  { aa=(a+k)->r_next, s=pp[k];
    while((aa=aa->r_next)!=anf) pp[aa->j]-= s*aa->x;
  }
  for (k=0; k<n; pp[k]/=((a+k)->r_next)->x, k++);
  l=n;
  while(--l)
  { for (k=0, s=pp[l]; k<l; k++)
    { aa=a+k; while((aa=aa->r_next)->j<l);
      if(aa->j==l) pp[aa->i]-=s*aa->x;
    }
  }
//
  ls_REAL alpha, rp=r*p, ry; y=A*p;

  while(!((s=p*y)<eps)&&(it++<rep))
  { alpha=rp/s, x+=alpha * p, r-=alpha*y;
// Loesen auf y
    y=r;
    for (k=0; k<n-1; k++)
    { aa=(a+k)->r_next, s=yy[k];
      while((aa=aa->r_next)!=anf) yy[aa->j]-= s*aa->x;
    }
    for (k=0; k<n; yy[k]/=((a+k)->r_next)->x, k++);
    l=n;
    while(--l)
    { for (k=0, s=yy[l]; k<l; k++)
      { aa=a+k; while((aa=aa->r_next)->j<l);
        if(aa->j==l) yy[aa->i]-=s*aa->x;
      }
    }
//
    ry=r*y, p=y+(ry/rp)*p, rp=ry; y=A*p;
  }
  b=r; return it;
}

ls_Vector sp_sCG_precond::residuum(const ls_Vector &x, const ls_Vector &b) const
{  ls_Vector r(0,"Residuum"); r=b-A*x; return r;}

const sp_sCG_precond& sp_sCG_precond::write_row(ostream &f) const
{ f<<name<<endl<<"{"<<endl;
  A.write_row(f), F.write_row(f); f<<"}"<<endl;
  return *this;
}

const sp_sCG_precond& sp_sCG_precond::write_column(ostream &f) const
{ f<<name<<endl<<"{"<<endl;
  A.write_column(f); F.write_column(f); f<<"}"<<endl;
  return *this;
}

sp_sCG_precond& sp_sCG_precond::read_row(istream &f)
{ static char *q1="sp_sCG_precond::read_row";
  char w[ls_len];
  if(*name)while(f>>w&&strcmp(w,name));
  else f>>name, *(A.name)=0, *(F.name)=0;
  if(!f.good()) throw ls_error(name, q1, object_not_found_);
  f>>w;
  if(strcmp(w, "{")) throw ls_error(name,q1,object_contents_error_);
  A.read_row(f),F.read_row(f);
  if(!(f>>w)||(*w!='}')) throw ls_error(name, q1, object_end_error_);
  rc=0; return *this;
}

sp_sCG_precond& sp_sCG_precond::read_column(istream &f)
{ static char *q1="sp_sCG_precond::read_column";
  char w[ls_len];
  if(*name)while(f>>w&&strcmp(w,name));
  else f>>name, *(A.name)=0, *(F.name)=0;
  if(!f.good()) throw ls_error(name, q1, object_not_found_);
  f>>w;
  if(strcmp(w, "{")) throw ls_error(name,q1,object_contents_error_);
  A.read_column(f), F.read_column(f);
  if(!(f>>w)||(*w!='}'))
    throw ls_error(name, q1, object_end_error_);
  rc=0; return *this;
}

const sp_sCG_precond& sp_sCG_precond::operator>>(char *nam) const
{ if(!*nam)
  { cout.setf(ios::fixed, ios::floatfield);
    cout.precision(int(sizeof(ls_REAL)*1.5));
    write_row(cout);
  }
  else
  { ofstream f(nam);
    f.setf(ios::scientific, ios::floatfield);
    f.precision(int(sizeof(ls_REAL)*1.5));
    write_row(f);
  }
  return *this;
}

sp_sCG_precond& sp_sCG_precond::operator<<(char *nam)
{ if(!*nam) read_row(cin); else{ ifstream f(nam); read_row(f);}
  return *this;
}
#endif
